@@ -1,6 +1,6 @@
 package Gtk2::Ex::HyperTextBuffer;
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use strict;
 use Gtk2;
@@ -176,19 +176,18 @@ Like C<insert_blocks()> but inserts at current cursor position.
 
 =cut
 
-my $_Verbatim = 0;
-
 sub insert_blocks_at_cursor {
 	# Make sure not to modify content of the reference !
 	# TODO: check default editable / region editable
 	my $self = shift;
-	
-	$self->signal_handlers_block_by_func(\&on_insert);
+	$self->signal_handlers_block_by_func($_)
+		for \&on_insert, \&on_apply_tag, \&on_remove_tag;
 	my $begin = $self->get_iter_at_mark( $self->get_insert )->get_offset;
 	$self->_insert_blocks(@_);
 	my $end = $self->get_iter_at_mark( $self->get_insert )->get_offset;
 	$self->_record_step('I', $begin, $end, @_);
-	$self->signal_handlers_unblock_by_func(\&on_insert);
+	$self->signal_handlers_unblock_by_func($_)
+		for \&on_insert, \&on_apply_tag, \&on_remove_tag;
 	return 1;
 }
 
@@ -197,18 +196,25 @@ sub _insert_blocks {
 
 	for (@blocks) {
 		unless (ref $_) {
+			my $text = $_;
+			$self->insert_at_cursor($text);
 			if (@{$self->{edit_mode_tags}}) {
-				my $iter = $self->get_iter_at_mark( $self->get_insert );
-				$self->insert_with_tags(
-					$iter, $_, @{$self->{edit_mode_tags}} );
+				my $end = $self->get_iter_at_mark(
+					$self->get_insert );
+				my $start = $end->copy;
+				$start->backward_chars(length $text);
+				$self->remove_all_tags($start, $end);
+				$self->apply_tag($_, $start, $end)
+					for @{$self->{edit_mode_tags}};
 			}
-			else { $self->insert_at_cursor($_) }
 		}
 		else {
 			my ($type, $meta) = @$_[0, 1];
 			my $tag;
 			if ($type eq 'image') {
-				$self->insert_image_from_file(@$meta{qw/src width height/});
+				my %data = %$meta; # copy
+				my @arg = delete @data{qw/src width height/};
+				$self->insert_image_from_file(@arg, \%data);
 				next;
 			}
 			elsif ($type eq 'link') {
@@ -218,11 +224,11 @@ sub _insert_blocks {
 			else { $tag = $self->get_tag_table->lookup($type) }
 			
 			if ($tag) {
-				$_Verbatim++ if $type eq 'Verbatim'; # FIXME
-				push @{$self->{edit_mode_tags}}, $tag;
+				#push @{$self->{edit_mode_tags}}, $tag;
+				@{$self->{edit_mode_tags}} = ($tag);
 				$self->_insert_blocks(@$_[2 .. $#$_]); # recurs
-				pop @{$self->{edit_mode_tags}};
-				$_Verbatim-- if $type eq 'Verbatim'; # FIXME
+				#pop @{$self->{edit_mode_tags}};
+				@{$self->{edit_mode_tags}} = ();
 			}
 			elsif ($type eq 'Page' or $type eq 'Para') {
 				if ($$meta{base}) {
@@ -239,10 +245,9 @@ sub _insert_blocks {
 			$self->insert_at_cursor("\n"x$lines) if $lines;
 		}
 	}
-
 }
 
-=item C<insert_image_from_file(FILENAME, WIDTH, HEIGHT, ITER)>
+=item C<insert_image_from_file(FILENAME, WIDTH, HEIGHT, DATA, ITER)>
 
 Insert an image in the buffer at the cursor
 or at ITER if ITER is specified.
@@ -251,12 +256,14 @@ WIDTH and HEIGHT are optional arguments. If both are set the image
 is resized to these dimensions. If one of both is zero or undefined
 the image is scaled with fixed ratio.
 
+DATA is an optional hash reference that is attached to the image object.
+
 Use C<insert_blocks()> if you want this action to be un-doable. 
 
 =cut
 
 sub insert_image_from_file {
-	my ($self, $src, $w, $h, $iter) = @_;
+	my ($self, $src, $w, $h, $data, $iter) = @_;
 	($w, $iter) = (undef, $w) if @_ == 3 and ref $w;
 	my $file = $self->{base} ? File::Spec->rel2abs($src, $self->{base}) : $src;
 	#warn "file: $src, base: $self->{base} => $file\n";
@@ -278,6 +285,7 @@ sub insert_image_from_file {
 	
 	$pixbuf->{image_src} = $src;
 	$pixbuf->{image_file} = $file;
+	$pixbuf->{image_data} = $data;
 	$pixbuf->{resized} = ($w || $h) ? [$w, $h] : 0;
 	$iter ||= $self->get_iter_at_mark( $self->get_insert );
 
@@ -332,6 +340,8 @@ sub get_parse_tree {
 					my $meta = { src => $pixbuf->{image_src} };
 					@$meta{qw/width height/} = @{$pixbuf->{resized}}
 						if $pixbuf->{resized};
+					%$meta = (%{$pixbuf->{image_data}}, %$meta)
+						if $pixbuf->{image_data};
 					push @{$stack[-1]}, ['image', $meta];
 				}
 				$image->forward_char;
