@@ -11,7 +11,7 @@ PrintToBrowser
 
 =head1 VERSION
 
-0.18
+0.23
 
 =head1 DESCRIPTION
 
@@ -64,7 +64,7 @@ sub on_TmpPrint {
 	return $app->error_dialog("Could not find template for PrintToBrowser")
 		unless defined $template;
 	$exporter->{_template} =  $template;
-	my ($rep, $page) = @{$app}{'repository', 'page'};
+	my ($rep, $page) = @{$app}{'notebook', 'page'};
 	my $tmp = Zim::Page->new($exporter, $page->name);
 	$tmp->set_source( Zim::File->new($file) );
 	$tmp->set_format('html');
